--[[
PansyLabs Timers. Refer to LICENSE and README.md for details
]]--

function PL_Timers:New(control)
  instance = {}
  instance.control = control
  setmetatable(instance, { __index = PL_Timers })
  instance:Initialise()
  return instance
end

function PL_Timers:Initialise()
  self:InitialisePools()
  self.timerData = {}  
  self.activeCharacter = false
  self.updates = 0
  self.currentAlarms = {}
  self.slotUsage =  {}
  self.renderList = {}
  self.gracePeriod = false
  EVENT_MANAGER:RegisterForEvent(PL_Timers.name .. "OnPlayerActivated", EVENT_PLAYER_ACTIVATED, function(...) self:OnPlayerActivated(...) end)
  EVENT_MANAGER:RegisterForEvent(PL_Timers.name .. "OnPlayerDectivated", EVENT_PLAYER_DEACTIVATED, function(...) self:OnPlayerDeactivated(...) end)
end

function PL_Timers:OnPlayerActivated()
  if (not PL_Timers.Settings.enabled) then return end
  -- Only do this once, else it fires after every loading screen. Also provides a convenient place
  -- to stash the char name.
  if (self.activeCharacter == false) then
  self.gracePeriod = true
  self.activeCharacter = GetUnitName("player")
    -- Nuke the AWT entries for the active character, allowing other code to treat it as external timers only.
    -- It will be re-populated by OnPlayerDeactivated.
    PL_Timers.AWT.data[self.activeCharacter] = nil
    local startupDelay = 1000
    if IsEnlightenedAvailableForCharacter() then
      startupDelay = 5500
    end
    zo_callLater(function()
      if (PL_Timers.Settings.enableTimersOnLogin) then
        self:SlashTimers()
      end
      self.gracePeriod = false
    end, startupDelay)
  end
  self.running = true
end

-- On player deactivate store a snapshot of  the timers for this character.
-- This is done even if settings for AWT related features are off as the settings are currently per character.
function PL_Timers:OnPlayerDeactivated()
  -- Settings may be absent if there was a scripting error during init, this prevents a spurious error dialogue
  -- and another reloadui roundtrip.
  if (not PL_Timers.Settings) or (not PL_Timers.Settings.enabled) then return end
  PL_Timers.AWT.data[self.activeCharacter] = self:GetTimersForCharacter(self.activeCharacter)
end

-- Throttle updates and invoke other behaviour.
function PL_Timers:OnUpdate()
  if (not self.running) then return end
  if (not PL_Timers.Settings.enabled) then return end
  self.updates = self.updates + 1
  if (self.updates % 5 == 0) then
    self:RefreshModel()
    self:UITick()
  end
end

-- Handles a tick of the UI.
function PL_Timers:UITick()
  self:PrepareAlarms()
  if self.gracePeriod then return end
  self:Render()
  -- Always update whatever is on the screen.
  for _, binding in pairs(self.timerPool:GetActiveObjects()) do
    self:RefreshTimer(binding)
  end
end

-- Prepare to render alarms and defunct timers
function PL_Timers:PrepareAlarms()
  if not self.dueAlarms then return end
  for key, index in pairs(self.dueAlarms) do
    if self.currentAlarms[key] == nil then
      self.currentAlarms[key] = true
      self:PrepareTimer(index)
    end
  end
  -- Defunct timers, if any.
  if self.defunctTimers then
    for _, timer in pairs(self.defunctTimers) do
      local craft, line, trait = self:BuildDisplayData(timer)
      PL_Timers.msg(string.format(PL_Timers.Strings.extTimerCompleted, timer.characterName, craft, line, trait), "CCCC33")
      self:PrepareInfobox(timer.characterName, string.format(PL_Timers.Strings.extTimerCompletedInfobox, craft, line, trait), {0.8, 0.8, 0.8, 1})
    end
    self.defunctTimers = {}
  end
end

-- Refresh a single timer (specified by binding), updating the displayed details (countdown, colour etc.)
function PL_Timers:RefreshTimer(binding)
  local timer = self.timerData[binding.timerIdx]
  if (binding.control ~= nil and binding.control:IsHidden()) then return end
  local craft, line, trait, remaining, color = self:BuildDisplayData(timer)
  local descStr = string.format("%s - %s (%s)", craft, line, trait)
  binding.control:GetNamedChild("DescLabel"):SetText(descStr)
  local time = binding.control:GetNamedChild("TimeLabel")
  time:SetText(remaining)
  local timeBd = time:GetNamedChild("Backdrop")
  timeBd:SetCenterColor(color.rgba[1], color.rgba[2], color.rgba[3], color.rgba[4])
  timeBd:SetEdgeColor(color.rgba[1], color.rgba[2], color.rgba[3], color.rgba[4])
end

-- Builds & returns a set of human-friendly values for a timer.
function PL_Timers:BuildDisplayData(timer)
  local lineName = GetSmithingResearchLineInfo(timer.craft, timer.line)
  -- trait desc available as unused second return here:
  local traitType, _ = GetSmithingResearchLineTraitInfo(timer.craft, timer.line, timer.trait)
  local traitName = GetString("SI_ITEMTRAITTYPE", traitType)
  local rs = FormatTimeSeconds(timer.remaining, TIME_FORMAT_STYLE_DESCRIPTIVE_SHORT_SHOW_ZERO_SECS, TIME_FORMAT_PRECISION_SECONDS, TIME_FORMAT_DIRECTION_DESCENDING)
  local color = nil
  if (timer.remaining <= PL_Timers.Settings.redTime) then
    color = PL_Timers.Settings.RAG.red
  elseif (timer.remaining < PL_Timers.Settings.amberTime) then
    color = PL_Timers.Settings.RAG.amber
  else
    color = PL_Timers.Settings.RAG.green
  end
  return PL_Timers.Crafts[timer.craft].name, lineName, traitName, rs, color
end

function PL_Timers:AddToRenderList(charName, key, entry)
  local renderList = self.renderList[charName] or {timers={}, info={}}
  table.insert(renderList[key], entry)
  self.renderList[charName] = renderList
end

function PL_Timers:PrepareTimers(charName)
  local indexes = self:GetTimerIndexesForCharacter(charName)
  for _, index in pairs(indexes) do
    self:PrepareTimer(index)
  end
end

-- Prepares a single timer (whole widget, timeline, etc.) ready to be rendered.
function PL_Timers:PrepareTimer(index)
  local binding, poolKey = self.timerPool:AcquireObject()
  binding.timerIdx = index
  binding.timeline = GetAnimationManager():CreateTimelineFromVirtual("PL_Timers_Show", binding.control)
  -- At the end of the animation free up the timer. Ideally this'd go in the pool factory func but 
  -- then the pool key is not available...
  binding.timeline:SetHandler("OnStop", function(...)
    self:TimerCompleted(binding)
    self.timerPool:ReleaseObject(poolKey)
  end)  
  self:AddToRenderList(self.timerData[index].characterName, "timers", binding)
end

-- Prepares all the unused slot infoboxes for a character.
-- TODO: Add support for cross-character slot usage too. Tricky due to the alert settings for these.
function PL_Timers:PrepareUnusedSlotWarnings(charName)
  if charName ~= self.activeCharacter then return end
  local details = self:GetUnusedSlotDetails(charName)
  for craftType, detail in pairs(details) do
    local msg = string.format(PL_Timers.Strings.notUsingAllTimers, PL_Timers.Crafts[craftType].name, detail.used, detail.max)
    local color = detail.used == 0 and PL_Timers.Settings.RAG.red.rgba or PL_Timers.Settings.RAG.amber.rgba
    self:PrepareInfobox(charName, msg, color)
  end
end

-- Prepares an info box for a character.
function PL_Timers:PrepareInfobox(charName, msg, color)
    local info, poolKey = self.infoPool:AcquireObject()
    info:GetNamedChild("Label"):SetText(msg)
    info:GetNamedChild("Label"):SetColor(color[1], color[2], color[3], color[4])
    local timeline = GetAnimationManager():CreateTimelineFromVirtual("PL_Timers_Show", info)
    timeline:SetHandler("OnStop", function(...)
      self.infoPool:ReleaseObject(poolKey)
    end)
    -- No binding for infos so just stash the timeline somewhere convenient.
    info.timeline = timeline
    self:AddToRenderList(charName, "info", info)
end

-- Pre-process all of the prepared UI artefacts 
function PL_Timers:BuildTimelineSequence()
  if not next(self.renderList, nil) then return false end
  local timelines = {}
  -- Do the active character first.
  self:BuildTimelineSequenceForCharacter(timelines, self.activeCharacter)
  -- Now do any others.
  for characterName, _ in pairs(self.renderList) do
    if (characterName ~= self.activeCharacter) then
      self:BuildTimelineSequenceForCharacter(timelines, characterName)
    end
  end
  return timelines
end

-- Pre-process all of the prepared UI artefacts, for a single character.
function PL_Timers:BuildTimelineSequenceForCharacter(timelines, characterName)
  local data = self.renderList[characterName]
  if not data then return end
  local hasTimers = (#data.timers > 0)
  local hasInfo = (#data.info > 0)
  if hasTimers or hasInfo then
    table.insert(timelines, self:PrepareCaptionTimeline(characterName))
  end
  if hasTimers then
    for _, binding in pairs(data.timers) do
      table.insert(timelines, binding.timeline)
    end
  end
  if hasInfo then
    for _, info in pairs(data.info) do
      table.insert(timelines, info.timeline)
    end
  end
end

-- Fires all the timelines.
function PL_Timers:Render()
  local timelines = self:BuildTimelineSequence()
  if (not timelines) or (#timelines == 0) then return end
  -- TODO: This is a hack around an unknown bug. The symptom is that the timers will not display 
  -- if the character starts a new line of research. Checking visibility and alpha on TLC and timers
  -- shows they should be visible. Toggling the TLC visibility off and back seems to resolve the problem.
  PL_Timers_TopLevelControl:ToggleHidden()
  PL_Timers_TopLevelControl:ToggleHidden()
  for _, timeline in pairs(timelines) do
    local slot = self:TakeSlot()
    ZO_PreHookHandler(timeline, "OnStop", function()
      self:ReleaseSlot()
    end)
    -- Offset all the animations in the timeline by some amount, for a staggered start.
    for i = 1, timeline:GetNumAnimations() do
      local anim = timeline:GetAnimation(i)
      anim:SetDuration(anim:GetDuration() + (slot * 100))
    end
    -- Assume the first animated control is common to all animations. Offset it vertically.
    local anim = timeline:GetFirstAnimation()
    local control = anim:GetAnimatedControl()
    control:ClearAnchors()
    control:SetAnchor(TOP, self.control, TOP, 0, slot * 55)
    -- The control is hidden again when it is released to the pool; see pool code.
    control:SetAlpha(0)
    control:SetHidden(false)
    timeline:PlayFromStart()
  end  
  PL_Timers.PlaySound(next(self.currentAlarms) == nil and "timers" or "alarm")
  self.renderList = {} 
end

function PL_Timers:PrepareCaptionTimeline(text)
  local caption, poolKey = self.captionPool:AcquireObject()
  caption:GetNamedChild("Label"):SetText(text)
  local tl = ANIMATION_MANAGER:CreateTimelineFromVirtual("PL_Timers_Show", caption)
  tl:SetHandler("OnStop", function(...)
    self.captionPool:ReleaseObject(poolKey)
  end)
  return tl
end

-- Called from the OnStop handler on the timer animation
function PL_Timers:TimerCompleted(binding)
  -- If this was an alarm, remove it from the list. Bit of a hoop jump, why not put the key on the binding too?
  local key = self.timerData[binding.timerIdx].key
  if (self.currentAlarms[key] ~= nil) then
    self.currentAlarms[key] = nil
  end
end

-- Handles /timers
function PL_Timers:SlashTimers()
  if (not PL_Timers.Settings.enabled) then return end
  if (self:PeekSlot() > 0) and not(self.gracePeriod) then return end
  -- Active character
  self:PrepareTimers(self.activeCharacter)
  self:PrepareUnusedSlotWarnings(self.activeCharacter)
  -- Any other character
  if (not PL_Timers.Settings.enableShowAllChars) then return end
  for character, _ in pairs(PL_Timers.AWT.data) do
    if (character ~= self.activeCharacter) then
      self:PrepareTimers(character)
      self:PrepareUnusedSlotWarnings(character)
    end
  end  
end

-- Initialise ALL the things - set up the addon instance and hook up slash commands.
function PL_Timers_OnInitialized(self)
  PL_TIMERS = PL_Timers:New(self)
  SLASH_COMMANDS["/timers"] = function(...) PL_TIMERS:SlashTimers(...) end
end
