--[[
PansyLabs Timers. Refer to LICENSE and README.md for details

This file contains data model stuff.
]]--

-- Handles a tick of the data model. Determine the state of all ongoing research.
function PL_Timers:RefreshModel()

  local timers = {}
  local slotUsage = {}
  local now = GetTimeStamp()
  ---[[
  for craftType, craftData in pairs(PL_Timers.Crafts) do
    -- Do not globally cache number of research slots, can change during play.
    local maxSlots = GetMaxSimultaneousSmithingResearch(craftType)
    slotUsage[craftType] = { max = maxSlots, used = 0 }
    for researchLineIdx = 1, craftData.lines do
      for traitIdx = 1, PL_Timers.numTraitItems do
        local duration, remaining = GetSmithingResearchLineTraitTimes(craftType, researchLineIdx, traitIdx)
        if (remaining ~= nil) then
          local timer = {}
          timer.craft = craftType
          timer.line = researchLineIdx
          timer.trait = traitIdx
          timer.duration = duration
          timer.remaining = remaining
          timer.timestamp = now
          timer.external = false
          timer.characterName = self.activeCharacter
          -- This key is used in ui binding pooling and for tracking alarms.
          timer.key = table.concat({timer.craft, timer.line, timer.trait}, "_")
          timers[#timers + 1] = timer
          slotUsage[craftType].used = slotUsage[craftType].used + 1
        end
      end
    end
  end
  --]]
  --[[ Test data:
  if #self.timerData == 0 then
    timers = {
      { craft = 1, line = 6, trait = 5, duration=9000, remaining=75, characterName="Daigra Caridwulf", external=false, key = "1_6_5"},
      { craft = 6, line = 2, trait = 2, duration=9000, remaining=920, characterName="Daigra Caridwulf", external=false, key = "6_2_2"},
      { craft = 2, line = 10, trait = 6, duration=9000, remaining=95, characterName="Daigra Caridwulf", external=false, key = "2_10_6"},
      { craft = 2, line = 6, trait = 5, duration=9000, remaining=110, characterName="Sedyya", external=true, key = "2_6_5"},
      { craft = 2, line = 3, trait = 1, duration=9000, remaining=955, characterName="Sedyya", external=true, key = "2_3_1"},
    }
    slotUsage = {
      [1] = { max = 2, used = 1 },
      [2] = { max = 2, used = 2 },
      [6] = { max = 2, used = 0 },
    }
  else
    timers = self.timerData
    if ((self.updates % 40) == 0) then
      for _, data in pairs(timers) do
        data.remaining = data.remaining - 1
      end
    end
  end
  --]]
  
  self.slotUsage = slotUsage
  
  -- Manually tick the clock for other characters' timers. Note that this always bases the calculated time 
  -- on the time from AWT, NOT the value from the previous calculation. Also note this data is not written
  -- back to AWT - only the active (deactivating) char's data is written back out.
  if (PL_Timers.AWT ~= nil) then
    for characterName, extTimersForChar in pairs(PL_Timers.AWT.data) do
      if (characterName ~= self.activeCharacter) then
        for index, extTimer in pairs(extTimersForChar) do
          local dt = GetDiffBetweenTimeStamps(now, extTimer.timestamp)
          local calculatedRemainder = extTimer.remaining - dt
          local timer = {}
          timer.craft = extTimer.craft
          timer.line = extTimer.line
          timer.trait = extTimer.trait
          timer.duration = extTimer.duration
          timer.remaining = calculatedRemainder
          -- The updated timestamp is not currently required, as this data is never written back to AWT.
          --timer.timestamp = now
          timer.characterName = characterName
          timer.external = true
          -- This key is used in ui binding pooling and for tracking alarms.
          timer.key = table.concat({characterName, timer.craft, timer.line, timer.trait}, "_")
          timers[#timers + 1] = timer
          -- Remove AWT entries when they expire.
          if (calculatedRemainder <= 0) then
            table.remove(PL_Timers.AWT.data[characterName], index)
          end
        end
      end
    end
  end
  
  -- Look for defunct timers. For the active character, only do this during the grace period so as to notify
  -- of research that completed while logged out. Otherwise, only do it for other characters. This is cos the
  -- game will already notify the active character when it completes some research.
  local defunctTimers = {}  
  for _, timer in pairs(timers) do
    if (timer.remaining <= 0) then
      if (self.gracePeriod and timer.characterName == self.activeCharacter) then
        -- Don't bother to go and find the same timer in AWT, expect AWT for the active char is nuked on login.
        table.insert(defunctTimers, timer)
      elseif (PL_Timers.Settings.enableAlarmAllChars and timer.characterName ~= self.activeCharacter) then 
        -- Only include expired timers for other chars. if alarms for other chars are on.
        -- This could be its own setting.
        table.insert(defunctTimers, timer)
        -- These are removed from AWT when the expire, by the tick code further up.
      end
    end
  end
  
  local oldTimerData = self.timerData
  self.timerData = timers
  self.defunctTimers = defunctTimers
  self:RefreshAlarms(oldTimerData)
end

-- Find timers that need to be alarms, populate dueAlarms.
function PL_Timers:RefreshAlarms(oldTimerData)
  if not PL_Timers.Settings.enableAlarms then return end
  local alarms = {}
  for _, t1 in pairs(oldTimerData) do
    -- We don't care about this timer if it's not for the active char, and AWT alarms are turned off.
    if (PL_Timers.Settings.enableAlarmAllChars or t1.characterName == self.activeCharacter) then
      for index, t2 in pairs(self.timerData) do
        if (t1.key == t2.key) then      
          -- These timers are paired.
          local a1 = PL_Timers.Settings.alarm1Time + PL_Timers.Settings.alarmLeadTime
          local a2 = PL_Timers.Settings.alarm2Time + PL_Timers.Settings.alarmLeadTime
          local a3 = PL_Timers.Settings.alarm3Time + PL_Timers.Settings.alarmLeadTime
          --if (t1.remaining ~= t2.remaining) then d("Time has passed for timer " .. t1.key .. " (" .. t1.remaining .. " != " .. t2.remaining .. " // " .. a1) end
          if (t1.remaining >= a1 and t2.remaining <= a1)
          or (t1.remaining >= a2 and t2.remaining <= a2)
          or (t1.remaining >= a3 and t2.remaining <= a3) then
            -- This timer is just about to go through an important milestone so show an 'alarm'.
            alarms[t1.key] = index
          end
        end
      end
    end
  end
  self.dueAlarms = alarms
end

function PL_Timers:GetTimersForCharacter(name)
  local timers = {}
  for index, timer in pairs(self.timerData) do
    if timer.characterName == name then
      timers[index] = timer
    end
  end
  return timers
end

function PL_Timers:GetTimerIndexesForCharacter(name)
  local indexes = {}
  for index, timer in pairs(self.timerData) do
    if timer.characterName == name then
      table.insert(indexes, index)
    end
  end
  return indexes
end

-- TODO: Support charactername. Requires slot usage and settings are available via AWT.
function PL_Timers:GetUnusedSlotDetails(charName)
  local details = {}
  for craftType, slots in pairs(self.slotUsage) do
    local setting = PL_Timers.Settings["unusedSlotWarning" .. craftType]
    if (slots.used < slots.max) then
      -- Only care about unused slots if settings dictate.
      if (setting == 2) or (slots.used == 0 and setting == 3) then
        details[craftType] = {
          used = slots.used,
          max = slots.max
        }
      end
    end
  end
  return details
end
